#include <DFPlayerMini_Fast.h>
#include <SoftwareSerial.h>

#include <Adafruit_BusIO_Register.h>
#include <Adafruit_SPIDevice.h>
#include <Adafruit_PN532.h>

#include "LowPower.h"
#include <arduino-timer.h>

#include <EEPROM.h>


#define DEBUG 0
#define DEBUG_MP3 false

#define VERSION 2.1


#define PN532_SCK  (6)
#define PN532_MISO (5)
#define PN532_MOSI (4)
#define PN532_SS   (3)

#define WAKEUP_PIN 2
#define MP3_POWER_PIN 8
#define PN532_POWER_PIN 9

#define CELL_PIN A0
#define RANDOM_PIN A3

#define DEFAULT_VOLUME 25

#define ADDR_VOLUME 0
#define ADDR_CB 1
#define ADDR_PAUSE 2
#define ADDR_WELCOME 3

#define CB_FOLDER 0

#define CB_BATT_1      1 // nutne potrebuji dobit
#define CB_BATT_2      2 // zacina dochazet stava
#define CB_WELCOME_1   3 // ahoj, co si dnes pustime?
#define CB_WELCOME_2   4 // kdo me to budi?
#define CB_WELCOME_3   5 // jeee, bude pohadka. Jakou si pustime?
#define CB_PLAYLIST_1  6 // nac takova kouzla, mas tu jen jednu pohadku
#define CB_PLAYLIST_2  7 // dnes to budou dve pohadky za sebou
#define CB_PLAYLIST_3  8 // tri pohadky? no snad taky neusnu, nez je vsechny dopovim
#define CB_PLAYLIST_4  9 // to asi nebudou pohadky co?
#define CB_PLAYLIST_5  10 // parada, prehrajeme si vsechno
#define CB_VOLUME_10   11 // budu potichoucku, jako myska
#define CB_VOLUME_15   12 // budeme si septat
#define CB_VOLUME_20   13 // jojo, ztisim se malicko
#define CB_VOLUME_25   14 // takto mluvim normalne, ani moc potichu, ani moc nahlas
#define CB_VOLUME_30   15 // juhuuu, poradne si zarveme. Ale vic uz to neumim.
#define CB_VOLUME_E    21 // hotovo
#define CB_VOLUME_M    17 // trochu uberu no
#define CB_VOLUME_ME   18 // potiseji to nejde
#define CB_VOLUME_P    19 // tak tedy hlasiteji no
#define CB_VOLUME_PE   20 // hlasiteji to nejde
#define CB_DONE        21 // hotovo
#define CB_CRAZY_BOX   22 // ahoj, jsem ukecana krabicka a od ted si s tebou budu povidat
#define CB_CRAZY_END   23 // tak ja moc kecam? tak od ted uz nereknu ani slovo
#define CB_BYE         24 // pa pa
#define CB_ERROR       25 // ajajaj

#if DEBUG
#define NFC_READ_TIMEOUT (10 * 1000UL)
#define TIMER_INTERVAL 1000
#define MAX_PAUSED_TIME (20 * 1000UL)
#else
#define NFC_READ_TIMEOUT (0.3 * 60 * 1000UL) // 18 seconds
#define TIMER_INTERVAL 1000
#define MAX_PAUSED_TIME (2 * 60 * 1000UL) // 2 minutes
#endif

SoftwareSerial mp3_serial(10, 11); // RX, TX

DFPlayerMini_Fast mp3;
Adafruit_PN532 nfc(PN532_SCK, PN532_MISO, PN532_MOSI, PN532_SS);

Timer<1, millis> timer;

unsigned long now = 0;
uint8_t eq = 0;
bool _run = true;
bool _playing = false;
bool _paused = false;
bool _read = true;
uint32_t n = 0;

int16_t n_tracks_in_folder = 0;
uint8_t n_track_played = 0;

const float batt_mvc = 4.9609375;
float batt_counts = 0;
float batt_mv = 0;

uint8_t volume = DEFAULT_VOLUME;
uint8_t crazy_box = 0;

uint32_t max_paused_t = MAX_PAUSED_TIME;

uint8_t w = 0;

uint8_t _idle = 0;

void(* reset_device) (void) = 0; 

void setup()
{
  pinMode(LED_BUILTIN, OUTPUT);
  pinMode(WAKEUP_PIN, INPUT);
  pinMode(MP3_POWER_PIN, OUTPUT);
  pinMode(PN532_POWER_PIN, OUTPUT);
  pinMode(PN532_SS, OUTPUT);

  digitalWrite(LED_BUILTIN, LOW);
  digitalWrite(MP3_POWER_PIN, HIGH);
  digitalWrite(PN532_POWER_PIN, HIGH);

  randomSeed(analogRead(RANDOM_PIN));

  delay(200);

  #if DEBUG
  Serial.begin(9600);
  Serial.println(F("Begin"));
  #endif

  volume = EEPROM.read(ADDR_VOLUME);
  crazy_box = EEPROM.read(ADDR_CB);
  max_paused_t = EEPROM.read(ADDR_PAUSE);
  max_paused_t = max_paused_t * 60 * 1000UL;
  w = EEPROM.read(ADDR_WELCOME);

  _idle = 0;

  #if DEBUG
  Serial.print(F("volume: ")); Serial.println(volume);
  Serial.print(F("crazy_box: ")); Serial.println(crazy_box);
  Serial.print(F("max_paused_t: ")); Serial.println(max_paused_t);
  #endif

  if ((volume < 10) || (volume > 30))
  {
    volume = DEFAULT_VOLUME;
  }

  if ((max_paused_t < 1 * 60 * 1000UL) || (max_paused_t > 60 * 60 * 1000UL))
  {
    max_paused_t = MAX_PAUSED_TIME;
  }

  #if DEBUG
  Serial.print(F("volume: ")); Serial.println(volume);
  Serial.print(F("crazy_box: ")); Serial.println(crazy_box);
  Serial.print(F("max_paused_t: ")); Serial.println(max_paused_t);
  #endif

  delay(200);

  mp3_serial.begin(9600);
  mp3.begin(mp3_serial, DEBUG_MP3, 2000);

  nfc.begin();

  uint32_t nfc_version = nfc.getFirmwareVersion();
  if (!nfc_version)
  {
    #if DEBUG
    Serial.println(F("Reset device, getting NFC version failed"));
    delay(500);
    #endif
        
    reset_device();
  }
  nfc.SAMConfig();

  #if DEBUG
  Serial.print(F("NFC version: "));
  Serial.println(nfc_version);
  #endif


  int16_t mp3_version = mp3.currentVersion();
  if (mp3_version == -1)
  {
    #if DEBUG
    Serial.println(F("Reset device, getting mp3 version failed"));
    delay(500);
    #endif
    
    reset_device();
  }

  #if DEBUG
  Serial.print(F("MP3 version: "));
  Serial.println(mp3_version);
  #endif

  #if DEBUG
  Serial.print(F("Set volume to: "));
  Serial.println(volume);
  #endif
  
  mp3.volume(volume);
  delay(100);

  mp3.EQSelect(eq);
  delay(100);

  batt_counts = analogRead(CELL_PIN);
  batt_mv = batt_counts * batt_mvc;

  #if DEBUG
  Serial.print(F("Battery voltage: "));
  Serial.println(batt_mv);
  #endif

  if (crazy_box == 1)
  {
    if (batt_mv < 3500)
    {
      if (batt_mv < 3300)
      {
        // nutne potrebuji dobit
        mp3.playFolder(CB_FOLDER, CB_BATT_1);
        delay(2000);
      }
      else
      {
        // zacina dochazet stava
        mp3.playFolder(CB_FOLDER, CB_BATT_2);
        delay(2000);
      }
    }
    else
    {
      // ahoj, co si dnes pustime?
      // kdo me to budi?
      // jeee, bude pohadka. Jakou si pustime?
      uint8_t welcome[] = {CB_WELCOME_1, CB_WELCOME_3, CB_WELCOME_2};
      w++;
      if (w > 2)
      {
        w = 0;
      }

      EEPROM.write(ADDR_WELCOME, w);
      
      mp3.playFolder(CB_FOLDER, welcome[w]);
      delay(3000);
    }
  }

  attachInterrupt(digitalPinToInterrupt(WAKEUP_PIN), wakeup, HIGH);

  timer.every(TIMER_INTERVAL, check_player);
}

void loop()
{
  timer.tick();
  
  if (_playing || _paused)
  {
    if ((unsigned long)(millis() - now) > 350)
    {
      now = millis();
      toggle_led();
    }
  }

  if (_read)
  {
    digitalWrite(LED_BUILTIN, LOW);

    uint8_t success;
    uint8_t uid[] = { 0, 0, 0, 0, 0, 0, 0 }; // Buffer to store the returned UID
    uint8_t uidLength; // Length of the UID (4 or 7 bytes depending on ISO14443A card type)
  
    success = nfc.readPassiveTargetID(PN532_MIFARE_ISO14443A, uid, &uidLength, NFC_READ_TIMEOUT);
  
    if (success)
    {
      #if DEBUG
      Serial.print(F("UID Length: "));
      Serial.print(uidLength, DEC);
      Serial.println(F(" bytes"));
      Serial.print(F("UID Value: "));
      nfc.PrintHex(uid, uidLength);
      Serial.println(F(""));
      #endif

      uint8_t data[4];
      success = nfc.ntag2xx_ReadPage(4, data);

      if (success)
      {
        nfc.PrintHex(data, 4);

        uint8_t func = data[0];
        uint8_t param = data[1];
        uint8_t folder = data[2];
        uint8_t track = data[3];

        #if DEBUG
        Serial.print(F("func: "));
        Serial.println(func);
        Serial.print(F("param: "));
        Serial.println(param);
        #endif

        switch (func)
        {
          // player mode
          case 1: // 0x01
            n_tracks_in_folder = mp3.numTracksInFolder(folder);
            n_track_played = n_tracks_in_folder - 1;
    
            #if DEBUG
            Serial.print(F("Number of tracks in folder: "));
            Serial.println(n_tracks_in_folder);
            Serial.println();
            Serial.print(F("Start playing track number "));
            Serial.print(track);
            Serial.print(F(" from folder number "));
            Serial.println(folder);
            #endif

            if (crazy_box == 1)
            {
              switch (n_tracks_in_folder)
              {
                case 1:
                  // nac takova kouzla, mas tu jen jednu pohadku
                  mp3.playFolder(CB_FOLDER, CB_PLAYLIST_1);
                  delay(3000);
                  break;
                case 2:
                  // dnes to budou dve pohadky za sebou
                  mp3.playFolder(CB_FOLDER, CB_PLAYLIST_2);
                  delay(3000);
                  break;
                case 3:
                  // tri pohadky? no snad taky neusnu, nez je vsechny dopovim
                  mp3.playFolder(CB_FOLDER, CB_PLAYLIST_3);
                  delay(3000);
                  break;
                case 4:
                  // to asi nebudou pohadky co?
                  mp3.playFolder(CB_FOLDER, CB_PLAYLIST_4);
                  delay(3000);
                  break;
                default:
                  // parada, prehrajeme si vsechno
                  mp3.playFolder(CB_FOLDER, CB_PLAYLIST_5);
                  delay(3000);
                  break;
              }
            }
            
            mp3.playFolder(folder, 1);
    
            now = 0;
            _read = false;
    
            disable_spi();
            break;
          case 5: // 0x05
            // cteni na pokracovani
            {
              #if DEBUG
              Serial.println(F("Book mode"));
              #endif
              
              int16_t ntif = mp3.numTracksInFolder(folder);
  
              if ((track + 1) <= ntif)
              {
                uint8_t data[] = {func, param, folder, (uint8_t)(track + 1)};
  
                success = nfc.ntag2xx_WritePage(4, data);
  
                if (success)
                {
                  mp3.playFolder(folder, track);
  
                  now = 0;
                  _read = false;
                  
                  disable_spi();
                }
                else
                {
                  
                }
              }
              else
              {
                uint8_t data[] = {func, param, folder, 1};
  
                success = nfc.ntag2xx_WritePage(4, data);
  
                if (success)
                {
                  mp3.playFolder(folder, track);
  
                  now = 0;
                  _read = false;
                  
                  disable_spi();
                }
                else
                {
                  
                }
              }
            }
            break;
          // set volume
          case 160: // 0xA0
            #if DEBUG
            Serial.println(F("Set volume to 10"));
            #endif
  
            if (volume != 10)
            {
              volume = 10;
              mp3.volume(volume);
              delay(20);
  
              #if DEBUG
              Serial.println(F("Write new value of volume to EEPROM"));
              #endif
  
              EEPROM.write(ADDR_VOLUME, 10);

              if (crazy_box == 1)
              {
                // budu potichoucku, jako myska
                mp3.playFolder(CB_FOLDER, CB_VOLUME_10);
              }
            }
            else
            {
              if (crazy_box == 1)
              {
                // staci rict jednou
                mp3.playFolder(CB_FOLDER, CB_VOLUME_E);
              }
            }
  
            delay(2000);
            break;
          case 161: // 0xA1
            #if DEBUG
            Serial.println(F("Set volume to 15"));
            #endif
  
            if (volume != 15)
            {
              volume = 15;
              mp3.volume(volume);
              delay(20);
  
              #if DEBUG
              Serial.println(F("Write new value of volume to EEPROM"));
              #endif
  
              EEPROM.write(ADDR_VOLUME, 15);

              if (crazy_box == 1)
              {
                // budeme si septat
                mp3.playFolder(CB_FOLDER, CB_VOLUME_15);
              }
            }
            else
            {
              if (crazy_box == 1)
              {
                // staci rict jednou
                mp3.playFolder(CB_FOLDER, CB_VOLUME_E);
              }
            }
  
            delay(2000);
            break;
          case 162: // 0xA2
            #if DEBUG
            Serial.println(F("Set volume to 20"));
            #endif
  
            if (volume != 20)
            {
              volume = 20;
              mp3.volume(volume);
              delay(20);
  
              #if DEBUG
              Serial.println(F("Write new value of volume to EEPROM"));
              #endif
  
              EEPROM.write(ADDR_VOLUME, 20);

              if (crazy_box == 1)
              {
                // jojo, ztisim se malicko
                mp3.playFolder(CB_FOLDER, CB_VOLUME_20);
              }
            }
            else
            {
              if (crazy_box == 1)
              {
                // staci rict jednou
                mp3.playFolder(CB_FOLDER, CB_VOLUME_E);
              }
            }
  
            delay(2000);
            break;
          case 163: // 0xA3
            #if DEBUG
            Serial.println(F("Set volume to 25"));
            #endif
  
            if (volume != 25)
            {
              volume = 25;
              mp3.volume(volume);
              delay(20);
  
              #if DEBUG
              Serial.println(F("Write new value of volume to EEPROM"));
              #endif
  
              EEPROM.write(ADDR_VOLUME, 25);

              if (crazy_box == 1)
              {
                // takto mluvim normalne, ani moc potichu, ani moc nahlas
                mp3.playFolder(CB_FOLDER, CB_VOLUME_25);
              }
            }
            else
            {
              if (crazy_box == 1)
              {
                // staci rict jednou
                mp3.playFolder(CB_FOLDER, CB_VOLUME_E);
              }
            }
  
            delay(2000);
            break;
          case 164: // 0xA4
            #if DEBUG
            Serial.println(F("Set volume to 30"));
            #endif
  
            if (volume != 30)
            {
              volume = 30;
              mp3.volume(volume);
              delay(20);
  
              #if DEBUG
              Serial.println(F("Write new value of volume to EEPROM"));
              #endif
  
              EEPROM.write(ADDR_VOLUME, 30);

              if (crazy_box == 1)
              {
                // juhuuu, poradne si zarveme. Ale vic uz to neumim.
                mp3.playFolder(CB_FOLDER, CB_VOLUME_30);
              }
            }
            else
            {
              if (crazy_box == 1)
              {
                // staci rict jednou
                mp3.playFolder(CB_FOLDER, CB_VOLUME_E);
              }
            }
  
            delay(2000);
            break;
          case 165: // 0xA5
            #if DEBUG
            Serial.println(F("Decrease volume"));
            Serial.print(F("volume: ")); Serial.println(volume);
            #endif
  
            if (volume > 10)
            {
              volume = volume - 5;
              mp3.volume(volume);
              delay(20);
  
              #if DEBUG
              Serial.println(F("Write new value of volume to EEPROM"));
              Serial.print(F("volume: ")); Serial.println(volume);
              #endif
  
              EEPROM.write(ADDR_VOLUME, volume);

              if (crazy_box == 1)
              {
                // trochu uberu no
                mp3.playFolder(CB_FOLDER, CB_VOLUME_M);
              }
            }
            else
            {
              if (crazy_box == 1)
              {
                // potiseji to nejde
                mp3.playFolder(CB_FOLDER, CB_VOLUME_ME);
              }
            }
  
            delay(2000);
            break;
          case 166: // 0xA6
            #if DEBUG
            Serial.println(F("Increase volume"));
            Serial.print(F("volume: ")); Serial.println(volume);
            #endif
  
            if (volume < 30)
            {
              volume = volume + 5;
              mp3.volume(volume);
              delay(20);
  
              #if DEBUG
              Serial.println(F("Write new value of volume to EEPROM"));
              Serial.print(F("volume: ")); Serial.println(volume);
              #endif
  
              EEPROM.write(ADDR_VOLUME, volume);

              if (crazy_box == 1)
              {
                // tak tedy hlasiteji no
                mp3.playFolder(CB_FOLDER, CB_VOLUME_P);
              }
            }
            else
            {
              if (crazy_box == 1)
              {
                // hlasiteji to nejde
                mp3.playFolder(CB_FOLDER, CB_VOLUME_PE);
              }
            }
  
            delay(2000);
            break;
          case 176: // 0xB0
            // zmenit delku pauzy

            max_paused_t = param * 60 * 1000UL;

            #if DEBUG
            Serial.println(F("Write new value of pause to EEPROM"));
            #endif

            EEPROM.write(ADDR_PAUSE, param);

            if (crazy_box == 1)
            {
              // hotovo
              mp3.playFolder(CB_FOLDER, CB_DONE);
            }

            delay(2000);
            break;
          // crazy box
          case 203: // 0xCB
            crazy_box = 1;

            #if DEBUG
            Serial.println(F("Write new value of creazy box to EEPROM"));
            #endif
  
            EEPROM.write(ADDR_CB, crazy_box);

            // ahoj, jsem ukecana krabicka a od ted si s tebou budu povidat
            mp3.playFolder(CB_FOLDER, CB_CRAZY_BOX);

            delay(3000);
            break;
          case 204: // 0xCC
            #if DEBUG
            Serial.println(F("Clear EEPROM"));
            #endif
  
            for (int i = 0 ; i < EEPROM.length() ; i++)
            {
              EEPROM.write(i, 0);
            }

            toggle_led();

            #if DEBUG
            Serial.println(F("Clear EEPROM done, reset"));
            delay(500);
            #endif

            digitalWrite(LED_BUILTIN, LOW);
            disable_spi();
            digitalWrite(MP3_POWER_PIN, LOW);
            
            _run = false;
            _paused = false;
        
            #if DEBUG
            Serial.println(F("Go to sleep"));
            delay(500);
            #endif
            
            LowPower.powerDown(SLEEP_FOREVER, ADC_OFF, BOD_OFF);
            break;
          // crazy end
          case 206: // 0xCE
            if (crazy_box == 1)
            {
              crazy_box = 0;
  
              #if DEBUG
              Serial.println(F("Write new value of creazy box to EEPROM"));
              #endif
    
              EEPROM.write(ADDR_CB, crazy_box);

              // tak ja moc kecam? tak od ted uz nereknu ani slovo
              mp3.playFolder(CB_FOLDER, CB_CRAZY_END);
            }

            delay(3000);
            break;

          default:
            #if DEBUG
            Serial.print(F("Start playing track number "));
            Serial.print(track);
            Serial.print(F(" from folder number "));
            Serial.println(folder);
            #endif
  
            mp3.playFolder(folder, track);
    
            now = 0;
            _read = false;
    
            disable_spi();
            break;
        }
      }
      else
      {
        #if DEBUG
        Serial.println(F("Reset device, NFC failed"));
        delay(500);
        #endif
        reset_device();
      }
    }
    else
    {
      if (crazy_box == 1)
      {
        // pa pa
        mp3.playFolder(CB_FOLDER, CB_BYE);
        delay(2000);
      }
              
      digitalWrite(LED_BUILTIN, LOW);
      disable_spi();
      digitalWrite(MP3_POWER_PIN, LOW);

      _run = false;
      _paused = false;
      _read = false;

      #if DEBUG
      Serial.println(F("Go to sleep"));
      delay(500);
      #endif

      LowPower.powerDown(SLEEP_FOREVER, ADC_OFF, BOD_OFF);
    }
  }
}

void toggle_led()
{
  digitalWrite(LED_BUILTIN, !digitalRead(LED_BUILTIN));
}

void wakeup()
{
  delay(300);

  if (_playing)
  {
    if (!_paused)
    {
      #if DEBUG
      Serial.println(F("Pause"));
      #endif

      _paused = true;
      n = 0;
      mp3.pause();
    }
  }
  else
  {
    if (_paused)
    {
      #if DEBUG
      Serial.println(F("Resume"));
      #endif

      _paused = false;
      n = 0;
      mp3.resume();
    }
  }

  if (!_run)
  {
    #if DEBUG
    Serial.println(F("Reset"));
    delay(500);
    #endif
    
    reset_device();
  }
}

bool check_player(void *)
{
  int16_t mp3_status = mp3.getStatus();

  #if DEBUG
  Serial.println();
  Serial.print(F("MP3 player status: "));
  Serial.println(mp3_status);
  #endif

  switch (mp3_status)
  {
    case 512: // idle
      _idle++;

      if (_idle >= 3)
      {
        _playing = false;
    
        if (!_read)
        {
          #if DEBUG
          Serial.println();
          Serial.print(F("n_tracks_in_folder: "));
          Serial.println(n_tracks_in_folder);
          Serial.print(F("n_track_played: "));
          Serial.println(n_track_played);
          #endif
            
          if ((n_tracks_in_folder > 0) && (n_track_played > 0) && (n_tracks_in_folder > n_track_played))
          {
            _idle = 0;
            n_track_played--;
            mp3.playNext();
          }
          else
          {
            if (_paused)
            {
              if (crazy_box == 1)
              {
                // pa pa
                delay(2000);
                mp3.playFolder(CB_FOLDER, CB_BYE);
                delay(2000);
              }
            }
            
            digitalWrite(LED_BUILTIN, LOW);
            disable_spi();
            digitalWrite(MP3_POWER_PIN, LOW);
        
            _run = false;
            _paused = false;
    
            #if DEBUG
            Serial.println(F("Go to sleep"));
            delay(500);
            #endif
        
            LowPower.powerDown(SLEEP_FOREVER, ADC_OFF, BOD_OFF);
          }
        }
      }
      break;
    case 513: // playing
      _playing = true;
      _paused = false;
      break;
    case 514: // paused
      _playing = false;
      _paused = true;
  
      n++;
  
      if (n >= (max_paused_t / TIMER_INTERVAL))
      {
        if (crazy_box == 1)
        {
          // pa pa
          mp3.playFolder(CB_FOLDER, CB_BYE);
          delay(2000);
        }
          
        digitalWrite(LED_BUILTIN, LOW);
        disable_spi();
        digitalWrite(MP3_POWER_PIN, LOW);
    
        _run = false;
        _paused = false;
  
        #if DEBUG
        Serial.println(F("Go to sleep"));
        delay(500);
        #endif
    
        LowPower.powerDown(SLEEP_FOREVER, ADC_OFF, BOD_OFF);
      }
      break;
    default:
      // -1 error
      // 0 

      if (crazy_box == 1)
      {
        // ajajaj
        mp3.playFolder(CB_FOLDER, CB_ERROR);
        delay(2000);
      }
          
      digitalWrite(LED_BUILTIN, LOW);
      disable_spi();
      digitalWrite(MP3_POWER_PIN, LOW);
      
      _run = false;
      _paused = false;
  
      #if DEBUG
      Serial.println(F("Go to sleep"));
      delay(500);
      #endif
      
      LowPower.powerDown(SLEEP_FOREVER, ADC_OFF, BOD_OFF);
      break;
  }

  return true;
}

void disable_spi()
{
  digitalWrite(PN532_POWER_PIN, LOW);

  pinMode(PN532_SCK, OUTPUT);
  pinMode(PN532_MOSI, OUTPUT);
  pinMode(PN532_SS, OUTPUT);

  digitalWrite(PN532_SCK, HIGH);
  digitalWrite(PN532_MOSI, HIGH);
  digitalWrite(PN532_SS, HIGH);
}
